
import numpy as np
import sys

from Utilities_LFC import compute_alpha_Tpnu # alpha_p, T_p(nu)


def LFC(start_time, end_time, tau, U0, dtU0, L_mat, g_func=None, \
        output=None, p=1, nu=1, startingValue="Pp_Ppprime"):
    """ Leapfrog-Chebyshev method 
    
    LFc method for the second order differential equation 
    u''(t) = - L_mat * u(t) - g(u(t)),    u(0) = U0,    u'(0) = dtU0
    
    Parameters
    ----------
    start_time : float
        Starting time of the LFC method.
    
    end_time : float
        Ending time of the LFC method; If the difference of `end_time` and 
        `start_time` divided by `tau` is not an integer, the end_time will not 
        be exactly reached.
    
    tau : float
        Time step size for the LFC method.
        
    U0, dtU0 : numpy ndarrays, 1-dimensional
        Initial values for u(0) and u'(0).
    
    L_mat : matrix_like, e.g. numpy ndarray, 2-dimensional
        Symmetric, positive definite matrix; L_mat can be any object, which
        has a method `dot` for matrix vector multiplication.
        The method `dot(U)` must have as parameter an 1-dimensional ndarray
        and return an 1-dimensional ndarray.
    
    g_func : function
        Function for the semilinearity of the differential equation. If None,
        the linear problem is solved, otherwise the semilinear problem.
        g_func needs as parameter an 1-dimensional ndarray (U) and must returns
        an 1-dimensional ndarray.
        
    output : objects as in file examples/Output.py
    
    p : int
        Polynomial degree for the LFC method, p must be an integer greater 
        than 1.
    
    nu : float
        Damping parameter for the LFC method, nu must be an number greater equal 
        than 1.
    
    startingValue : {'Pp_Ppprime', 'Taylor_3', 'Taylor_5', 'exact'}
        Method for the first time step; For 'Taylor_5' the semilinear case 
        is not implemented, for 'exact' the values in `dtU0` are taken as 
        solution for the first time step.
    
    Returns
    -------
    Un : ndarray, 1-dimensional
        Numerical solution of LFC method at time `time` (approximately).
    
    time : float
        Time at the end of the LFC method, cf. `end_time`.
    
    """

    # 
    time = float(start_time)
    # number of time steps 
    num_timesteps = int(np.round((end_time-start_time)/tau))
    
    end_time_real = start_time + num_timesteps*tau
    # print warning , if end_time cannot be exactly reached with time step size
    if (np.abs(end_time_real-end_time) > 1e-12):
        print("\n---Warning---")
        print("The given ending time={0:6.4f} will not be exactly equal to the"\
              " computed one ({1:6.4f}).\n".format(end_time, end_time_real))
    
    # constants for computation of Pp and Ppprime
    alpha, Tpnu = compute_alpha_Tpnu(p, nu)
    

    ## initial values with different approximations for first step
    # u(start_time)
    Unm = U0.copy()  
    
    # u(start_time + tau) 
    if startingValue == "Taylor_3": # Taylor order 3
        Un = L_mat.dot(U0)
        Un *= -0.5*tau**2
        Un += U0 + tau*dtU0
        if (not g_func is None):
            Un -= (0.5*tau**2) * g_func(U0)
    elif startingValue == "Taylor_5": # Taylor order 5
        Un = U0 + tau * dtU0
        u_1 = (-1./24*tau**4) * (L_mat.dot(U0))
        u_1 -= (0.5*tau**2) * U0 + (1./6*tau**3) * dtU0
        Un += L_mat.dot(u_1)
        if (not g_func is None):
            print("\n---ERROR---\nTaylor_5 not implemented in semilinear case")
            sys.exit(1)
    elif startingValue == "Pp_Ppprime": # Pp and Pp'
        Un = U0.copy()
        Un -= (0.5)*computePp(L_mat, U0, tau, p, nu, alpha, Tpnu)
        Un += tau*computePpprime(L_mat, dtU0, tau, p, nu, alpha, Tpnu)
        if (not g_func is None):
            Un -= (0.5*tau**2) * g_func(U0)
    elif startingValue == "exact":
        Un = dtU0
    else: 
        print("\n---ERROR---\nillegal choice of StartingValue")
        sys.exit(1)
    
    # advance time
    time += tau
    
    #output
    if (not output is None):
        output.compute_initial_output(Unm, Un, time, tau)
    
    
    unstable = False
    ## loop over time steps
    for k in range(1,num_timesteps):
        
        # update solution
        Unp = (-1)*computePp(L_mat, Un, tau, p, nu, alpha, Tpnu)
        if (not g_func is None):
            Unp -= (tau**2) * g_func(Un)
        Unp += 2*Un - Unm
            
        # update time
        time += tau
        
        # output
        if (not output is None):
            unstable = output.compute_output(Un, Unp, time, tau, k)
        
        # if unstable, stop iterating 
        if unstable:
            break
        
        # set for next time step
        Unm = Un
        Un = Unp
        
    # end loop over time steps
    
    # return numerical solution at time time (not necessarily equal to end_time)
    return Un, time


def computePp(L_mat, U, tau, deg_p, nu_p, alpha, Tpnu):
    """ Computation of Pp(Z)*U, Z = tau**2 * L_mat, via Algorithm 6.1 
    for polynomial degree deg_p, damping parameter nu, and values alpha and Tpnu 
    given through function compute_alpha_Tpnu(deg_p, nu_p)
    """

    if deg_p==1:
        # P1(Z)*U = Z*U
        U_out = L_mat.dot(U)
        U_out *= (tau**2)
        return U_out
    else:
        # initialize recursion
        tPkm2 = np.zeros(U.size)
        tPkm1 = L_mat.dot(U)
        tPkm1 *= 2*(tau**2)/(alpha*nu_p)
        tPk=None

        for k in range(2,deg_p+1):
            tPk = L_mat.dot(2*U - tPkm1)
            tPk *= (tau**2)/alpha
            tPk += nu_p*tPkm1
            tPk *= 2*Tpnu[k-1]/Tpnu[k]
            tPk -= (Tpnu[k-2]/Tpnu[k])*tPkm2

            if k < deg_p:
                tPkm2 = tPkm1
                tPkm1 = tPk

        return tPk


def computePpprime(L_mat, U, tau, deg_p, nu_p, alpha, Tpnu):
    """ Computation of Pp'(Z)*U, Z = tau**2 * L_mat, 
    for polynomial degree deg_p, damping parameter nu, and values alpha and Tpnu 
    given through function compute_alpha_Tpnu(deg_p, nu_p)
    """

    p_U = deg_p-1

    if deg_p==1:
        # P1'(Z) = 1
        U_out = np.copy(U)
        return U_out
    elif deg_p==2:
        # P2'(Z) = 8/(alpha*T_2(nu)) * (nu - Z/alpha)
        U_out = L_mat.dot(U)
        U_out *= -(tau**2)/alpha
        U_out += nu_p*U
        U_out *= 8/(alpha*Tpnu[deg_p])
        return U_out
    else:
        # Pp'(Z) = 2*p/(alpha*T_p(nu))*U_{p-1}(nu - Z/alpha)
        # U_p(x) = 2*x*U_{p-1}(x) - U_{p-2}(x),  U_1(x)=2x,  U_0(x)=1
        
        # initialize recursion
        Upm2 = U
        Upm1 = L_mat.dot(U)
        Upm1 *= -(tau**2)/alpha
        Upm1 += nu_p*U
        Upm1 *= 2
        Up = None

        for k in range(2,p_U+1):
            Up = L_mat.dot(Upm1)
            Up *= -(tau**2)/alpha
            Up += nu_p*Upm1
            Up *= 2
            Up -= Upm2
            
            if k < p_U:
                Upm2 = Upm1
                Upm1 = Up
            
        return (2*deg_p/(alpha*Tpnu[deg_p]))*Up      
