
import numpy as np
#import sys

def nu_order_four(deg_p):
    """ Return the optimal value for nu_p for a given polynomial degree
    deg_p (<=10) such that the LFC scheme is of order four"""
    
    nu_p = None
    if deg_p==1:
        print("\n---Warning---")
        print("For p=1 nu cannot be chosen such that we have order 4.",
            "nu is set to 1.\n")
        #sys.exit(1)
    elif deg_p==2:
        nu_p = 1.2247448713915890490986420373529 # sqrt(6)/2
    elif deg_p==3:
        nu_p = 1.0290855136357461251609905237902 # sqrt(sqrt(5)/4 + 1/2) 
    elif deg_p==4:
        nu_p = 1.0082607497358431721150033851633
    elif deg_p==5:
        nu_p = 1.0032332578774403572516531103429
    elif deg_p==6:
        nu_p = 1.001522817775525122588373121359
    elif deg_p==7:
        nu_p = 1.0008106154589539645479324497267
    elif deg_p==8:
        nu_p = 1.0004709549496960463813214408508
    elif deg_p==9:
        nu_p = 1.000292241979967210191208691395
    elif deg_p==10:
        nu_p = 1.0001909176571096286303911869917
    else:
        print("\n---Warning---")
        print("For this value of p={0:6.3f} no optimal nu is provided. nu is set to 1.\n"
            .format(deg_p))
        #sys.exit(1)      
    
    return nu_p


def LFC_constants(deg_p, nu_p):
    """ Computation of alpha, beta, m1, m2, m3 and T_k(nu_p) for given polynomial 
    degree deg_p and damping parameter nu_p"""
    
    # Chebyshev polynomials and derivatives
    coef = np.zeros(deg_p+1)
    coef[deg_p] = 1
    T_p = np.polynomial.chebyshev.Chebyshev(coef)
    dxT = T_p.deriv()
    dxxT = dxT.deriv()
    
    # constants
    alpha = 2.*dxT(nu_p)/T_p(nu_p)
    beta = alpha*(nu_p+1)
    m1 = 0.5*(1-1./T_p(nu_p))
    m2 = 4.*m1/beta
    m3 = T_p(nu_p)*dxxT(nu_p)/(4.*(dxT(nu_p)**2))
    
    Tpnu = np.zeros(deg_p)    
    for k in range(deg_p+1):
        coef = np.zeros(k+1)
        coef[k] = 1
        T_p = np.polynomial.chebyshev.Chebyshev(coef)
        Tpnu[k] = T_p(nu_p)

    return alpha, beta, m1, m2, m3, Tpnu


def compute_alpha_Tpnu(deg_p, nu_p):
    """ Computation of alpha_p and T_k(nu_p) for k=0,1,...,deg_p (with Chebyshev
    recursion of first and second kind) for given polynomial degree deg_p and 
    damping parameter nu_p"""
    
    alpha = None
    Tpnu = None
    
    if nu_p == 1:
        Tpnu = np.ones(deg_p+1)
        alpha = 2.*deg_p**2;
        
    else:
        # initialization 
        Tpnu = np.zeros(deg_p+1)
        Upnu = np.zeros(deg_p+1)

        Tpnu[0] = 1.
        Tpnu[1] = nu_p
        Upnu[0] = 1.
        Upnu[1] = 2*nu_p

        # recursion
        for k in range(2,deg_p+1):
            Tpnu[k] = 2*nu_p*Tpnu[k-1] - Tpnu[k-2]
            Upnu[k] = 2*nu_p*Upnu[k-1] - Upnu[k-2]

        alpha = 2*deg_p * Upnu[deg_p-1]/Tpnu[deg_p]

    return alpha, Tpnu

