 
import numpy as np

# for plot of the triangulation
import matplotlib.pyplot as plt
import matplotlib.tri as tri


class Error_Energy_HO:
    """ Class for computation of energy and error for the harmonic oscillator
    """
    
    def __init__(self, u_exact, omega2, num_steps):
        self.u_exact = u_exact
        self.omega2 = omega2
        
        ## initialize output vectors for energies
        self.energy_L2 = np.zeros(num_steps+1)# measured in discrete L2 norm
        self.energy = np.zeros(num_steps)     # measured in energy norm

    def compute_initial_output(self, u0, u1, time, tau):
    
        # initial values: energy
        self.energy_L2[0] = np.abs(u0)
        self.energy_L2[1] = np.abs(u1)
        self.energy[0] = np.sqrt( (1./tau**2)*(u1-u0)**2 \
                            + 0.25*self.omega2*(u1+u0)**2 )

    def compute_output(self, un, unp, time, tau, timestep):

        # energy    
        self.energy_L2[timestep+1] = np.abs(unp)
        self.energy[timestep] = np.sqrt( (1./tau**2)*(unp-un)**2 \
                            + 0.25*self.omega2*(unp+un)**2 )

        # test if method is unstable
        # if unstable, set output to 1e10 and return False
        if (self.energy[timestep]) > 10000:
            self.energy[timestep:] = 1e10
            self.energy_L2[timestep+1:] = 1e10
            return True
        else:
            return False

    def return_output(self):
        return self.energy_L2,  self.energy


class Error_Wave:
    """ Class for computation of error in energy norm for the linear acoustic 
    wave equation
    """
    
    def __init__(self, u_exact, M, A):
        self.Uxy = u_exact[0]
        self.ut = u_exact[1]
        self.M = M
        self.A = A
        
        ## initialize output vectors for error measured in energy norm
        self.error = 1      
        self.En = None

    def compute_initial_output(self, U0, U1, time, tau):
        
        # initial values: error
        En = U0-self.Uxy*self.ut(time-tau)
        Enp = U1-self.Uxy*self.ut(time)
        self.error = np.sqrt( (1./tau**2)*((Enp-En).dot(self.M.dot(Enp-En))) \
                            + 0.25*((Enp+En).dot(self.A.dot(Enp+En))) )

        # save for next time step
        self.En = Enp

    def compute_output(self, Un, Unp, time, tau, timestep):
        
        Enp = Unp-self.Uxy*self.ut(time)
        # error
        err_new = np.sqrt( (1./tau**2)*((Enp-self.En).dot(self.M.dot(Enp-self.En))) \
                        + 0.25*((Enp+self.En).dot(self.A.dot(Enp+self.En))) )        
        
        if self.error < err_new:
            self.error = err_new
        
        self.En = Enp
        
        # test if method is unstable
        # if unstable, set output to 1e10 and return False
        if (self.error) > 10000:
            self.error = 1e10
            
            return True
        else:
            return False
    
    def return_output(self):
        return self.error


def plot_tria(tria):
    """ Plot triangulation"""
    
    triang = tri.Triangulation(tria['x'], tria['y'], tria['triangles'])
    
    plt.figure()
    plt.triplot(triang, linewidth=0.5, color='k')
    plt.title("Triangulation")
    plt.show()
