
"""
Example 1
LFC method applied to the harmonic oscillator with different starting values and
no damping. For more information see figure 7.1 in 
'On leap-frog-Chebyshev schemes'.
"""

print("Example 1\nLFC method applied to the harmonic oscillator \
with different starting values and no damping\n")

import numpy as np
from matplotlib import pyplot as plt

## Add the parent folder path to the sys.path list
import sys
from pathlib import Path
sys.path.append(str(Path('.').absolute().parent))

import LFCmethod
from Output import Error_Energy_HO

#%% general data

# initial values
q0 = np.array([2.])
dtq0 = np.array([1.])

# frequency
omega = np.array([2.])
omega2 = omega**2

# starting time
start_time = 0

# number of time steps
num_steps = np.array([5, 10, 15, 20])

# LFC parameters
p = 5 # polynomial degree of P_p(z), integer, >=1
nu = 1. # damping parameter, float, >=1


#%% initialization

# exact solution of harmonic oscillator
qex = lambda t: np.cos(omega*(t-start_time))*q0 \
      + 1./omega*np.sin(omega*(t-start_time))*dtq0

# timesteps for which Pp(z)=0 or Pp(z)=4, z=tau^2 omega^2 (only for nu=1)
crit_tau = np.sqrt(2*(p**2)*(1 - np.cos(np.arange(p+1)*np.pi/p)) / omega2)

# time step sizes
tau_cfl = np.max(crit_tau)
tau = np.linspace(0.001, 1, 1000)*tau_cfl

# maximum of absolute vale of solution for different choices of q1
max_sol_Pp = np.zeros([tau.size, num_steps.size])
max_sol_Tay = np.zeros([tau.size, num_steps.size])

#%% time integration for different timesteps and number of timesteps

for kk in range(num_steps.size):

    startingValue = "Pp_Ppprime"
    for k in range(tau.size):
        err_en_Pp = Error_Energy_HO(qex, omega2, num_steps[kk])
        qn, time = LFCmethod.LFC(start_time, num_steps[kk]*tau[k], tau[k], \
                            q0, dtq0, omega2, None, err_en_Pp, p, nu, startingValue)

        # energies and errors
        en_L2, en = err_en_Pp.return_output()
        max_sol_Pp[k,kk] = np.max(en_L2)

    startingValue = "Taylor_3"
    for k in range(tau.size):
        err_en_Tay = Error_Energy_HO(qex, omega2, num_steps[kk])
        qn, time = LFCmethod.LFC(start_time, num_steps[kk]*tau[k], tau[k], \
                            q0, dtq0, omega2, None, err_en_Tay, p, nu, startingValue)

        # energies and errors
        en_L2, en = err_en_Tay.return_output()
        max_sol_Tay[k,kk] = np.max(en_L2)


#%% Plot
# Plot of maximum of absolute value of solution (values in max_sol__Pp and
# max_sol_Tay) for different timesteps and number of timesteps

plt.figure()

# Starting value Pp_Ppprime
ax1=plt.subplot(1, 2, 1)
for  kk in range(num_steps.size):
    plt.plot(tau**2*omega2, np.fliplr(max_sol_Pp)[:,kk], '.-',
             label="N={0:3d}".format(num_steps[-1-kk]))
handles, labels = ax1.get_legend_handles_labels()
plt.legend(handles[::-1], labels[::-1])

# values for which Pp(z)=0 or Pp(z)=4, z=tau^2 omega^2 (only for nu=1)
for k in crit_tau[1:]:
    plt.plot([k**2*omega2, k**2*omega2], [0, 10000], 'c:')

plt.xlim([0, 102])
plt.ylim([0, 35])
plt.xlabel(r"$\tau^2 \omega^2$")
plt.ylabel(r"$\max_n \vert u_n \vert$")
plt.title("Starting value Pp_Ppprime")

# Starting value Taylor5
ax2=plt.subplot(1, 2, 2)
for  kk in range(num_steps.size):
    plt.plot(tau**2*omega2, np.fliplr(max_sol_Tay)[:,kk], '.-',
             label="N={0:3d}".format(num_steps[-1-kk]))
plt.legend()
handles, labels = ax2.get_legend_handles_labels()
plt.legend(handles[::-1], labels[::-1])

# values for which Pp(z)=0 or Pp(z)=4, z=tau^2 omega^2 (only for nu=1)
for k in crit_tau[1:]:
    plt.plot([k**2*omega2, k**2*omega2], [0, 10000], 'c:')

plt.xlim([0, 102])
plt.ylim([0, 3500])
plt.xlabel(r"$\tau^2 \omega^2$")
plt.ylabel(r"$\max_n \vert u_n \vert$")
plt.title("Starting value Taylor5")

plt.show()
