
"""
Example 2
LFC method applied to the semilinear acoustic wave equation with different 
values for the damping parameter nu, compared to leapfrog method.
For more information see figure 7.3 (c) in 
'On leap-frog-Chebyshev schemes'.
"""

print("Example 2\nLFC method applied to the semilinear acoustic wave equation \
with different values for the damping parameter nu, \
compared to leapfrog method\n")

import numpy as np
import scipy.sparse as scp
import scipy.sparse.linalg as scplin
from matplotlib import pyplot as plt

## Add the parent folder path to the sys.path list
import sys
from pathlib import Path
sys.path.append(str(Path('.').absolute().parent))

import LFCmethod
from Utilities_LFC import nu_order_four
from Output import Error_Wave
from Output import plot_tria

#%% general data

# LFC parameters
p = 5
# the last value of nu is chosen such that the LFC method is of order 4
nu = np.array([1.0, 1.00025, nu_order_four(p)]) 

# time integration parameters
start_time = 0.
end_time = 4.2

# order of Taylor approximation for initial value
startingValue = "Pp_Ppprime"

# semilinearity
d = 50 
def g_func(u_vec):
    return d*u_vec

# space discretization for the acoustic wave equation with interior penalty dG,
# see for example the book of Di Pietro, Ern: Section 4.2;
# parameters: degree_DG = 3, eta_stab = 0.85;
# for the triangulation uncomment the following line
#plot_tria(np.load("triang_ex2.npz"))

spmat_A = scp.load_npz("spmat_A_ex2.npz") # laplacian
spmat_M = scp.load_npz("spmat_M_ex2.npz") # mass matrix


#%% time step sizes

# compute largest eigenvalue of L = M^-1 A
lam_max = scplin.eigsh(spmat_A, 1, spmat_M, maxiter=1000, tol=1e-6, 
                return_eigenvectors=False)

# maximal stable time step for leapfrog method
tau_CFL_LF = 2./np.sqrt(lam_max)

tau_pow = np.linspace(0.99, 1.13, 15)
tau_LF = tau_CFL_LF**tau_pow
tau_LFC = np.array([0.02189866, 0.02016504, 0.01975352, 0.0193504, 0.01856867,
                    0.01709868, 0.01449859, 0.01229389, 0.01042444, 0.00883926,
                    0.00749513, 0.00538897, 0.00387465, 0.00278586, 0.00200302])


#%% exact solution and its derivative

# we assume u(t,x,y) = uxy(x,y)*ut(t)
# here: uxy(x,y) = sin(k1*pi*x)*sin(k2*pi*y])
k1 = 1.    # wave number x
k2 = 1.    # wave number y
# load projection of spatial part of exact solution from file
Uxy = np.load("Uxy_ex2.npz")['uxy'] 

omega = np.sqrt(2 * np.pi**2 + d)
ut = lambda t: np.cos(omega*t) + np.sin(omega*t)
dtut = lambda t: omega*(np.cos(omega*t) - np.sin(omega*t))


#%%
class Mm1A:
    def __init__(self, M, A):
        self.M = M
        self.A = A
        
        # LU decomposition of mass matrix M
        self.solverM = scplin.splu(self.M)
#        self.solverM = scplin.factorized(self.M)
    
    def dot(self, u_vec):
        return self.solverM.solve(self.A.dot(u_vec))
#        return self.solverM(self.A.dot(u_vec))

L_mat = Mm1A(spmat_M,spmat_A)

#%%
# initial values
U0 = Uxy*ut(start_time)
dtU0 = Uxy*dtut(start_time)

# array for error in energy norm
error_LFC = np.zeros([nu.size, tau_LFC.size])
error_LF = np.zeros([tau_LF.size])

# time integration
for kk in range(nu.size):
    
    for k in range(tau_LFC.size):
        err_wave = Error_Wave([Uxy, ut], spmat_M, spmat_A)
        Un, time = LFCmethod.LFC(start_time, end_time, tau_LFC[k], \
                            U0, dtU0, L_mat, g_func, err_wave, p, nu[kk], \
                            startingValue)
        
        print("LFC: Run with nu=" + str(nu[kk]) + " and tau=" + str(tau_LFC[k]))
        err = err_wave.return_output()
        error_LFC[kk,k] = err


# comparison with leapfrog method
for k in range(tau_LF.size):
    err_wave = Error_Wave([Uxy, ut], spmat_M, spmat_A)
    Un, time = LFCmethod.LFC(start_time, end_time, tau_LF[k], \
                             U0, dtU0, L_mat, g_func, err_wave)
    
    print("LF: Run with tau=" + str(tau_LF[k]))
    err = err_wave.return_output()
    error_LF[k] = err
    
    
#%% error plot
plt.figure()

# Starting value Pp_Ppprime
ax1=plt.subplot(1, 2, 1)
for kk in range(nu.size):
    plt.loglog(tau_LFC, error_LFC[kk,:], '.-',
             label="LFC5, nu={0:3.6e}".format(nu[kk]))
#handles, labels = ax1.get_legend_handles_labels()
#plt.legend(handles[::-1], labels[::-1])
plt.loglog(tau_LF, error_LF, '.-',label="LF")
plt.legend()

# reference line
plt.loglog(tau_LFC, 4e2*tau_LFC**2, 'k--')
#plt.loglog(tau_LFC, 1.5e3*tau_LFC**4, 'k--')

# multiples of the maximal stable time step size for the leapfrog method
for k in range(p):
    plt.plot([tau_CFL_LF*(k+1), tau_CFL_LF*(k+1)], [0, 10000], 'k:')
    
plt.xlim([tau_LFC[-1]*0.8, tau_LFC[0]*1.1])
plt.ylim([1e-3, 1e0])
plt.xlabel(r"$\tau$")
plt.ylabel(r"$\max _n \vert\!\Vert e_{n+1/2} \Vert\!\vert$")
plt.title("Comparison of LFC5 and leapfrog method")

plt.show()

